@@ -192,18 +192,6 @@ class Batch(Workflow, ModelSQL, ModelView):
             default = default.copy()
         default.setdefault('state', cls.default_state())
         default.setdefault('lines', None)
-
-        # should we add copies of lines?
-        #new_batches = []
-        #for batch in batches:
-        #    default['name'] = batch['name'] + ' copy'
-        #    new_batch, = super(cls, cls).copy([batch], default=default)
-        #    print(new_batch)
-        #    Line.copy(batch.lines, default={
-        #            'batch': new_batch.id,
-        #            })
-        #    new_batches.append(new_batch)
-        #return new_batches
         return super(Batch, cls).copy(batches, default=default)
 
     @classmethod
@@ -364,7 +352,7 @@ class BatchLine(ModelSQL, ModelView):
     def __setup__(cls):
         super(BatchLine, cls).__setup__()
         # batch is in excludes to avoid general update of all moves.
-        cls._check_modify_exclude = {'batch', 'sequence'}
+        cls._check_modify_exclude = {'code', 'batch', 'sequence'}
         cls._order[0] = ('id', 'DESC')
 
     def get_rec_name(self, name):
@@ -595,7 +583,7 @@ class BatchLine(ModelSQL, ModelView):
                 self.contra_account = self.invoice.account.id
 
     @fields.depends('date', 'fiscalyear', 'journal', 'account',
-        'contra_account')
+        'contra_account', 'batch')
     def on_change_date(self):
         pool = Pool()
         Fiscalyear = pool.get('account.fiscalyear')
@@ -697,6 +685,10 @@ class BatchLine(ModelSQL, ModelView):
         pool = Pool()
         Move = pool.get('account.move')
         MoveLine = pool.get('account.move.line')
+        Invoice = pool.get('account.invoice')
+
+        _update_tax_cash_basis = getattr(
+            Invoice, '_update_tax_cash_basis', None)
 
         Move.post([l.move for l in lines
                 if l.move and l.state == 'draft'])
@@ -729,6 +721,10 @@ class BatchLine(ModelSQL, ModelView):
                         line.invoice.get_reconcile_lines_for_amount(_ZERO)
                     if remainder == _ZERO and move_line in reconcile_lines:
                         MoveLine.reconcile(reconcile_lines)
+            # Compatibility with account_tax_cash
+            if _update_tax_cash_basis:
+                with Transaction().set_context(payment_date=line.date):
+                    Invoice._update_tax_cash_basis([line.invoice])
 
     @classmethod
     def check_unposted_batch_lines_for_account(cls, account_ids=None):
@@ -756,18 +752,23 @@ class BatchLine(ModelSQL, ModelView):
         pool = Pool()
         Sequence = pool.get('ir.sequence')
 
-        vlist = [x.copy() for x in vlist]
-        for vals in vlist:
-            if not vals.get('sequence'):
-                vals['sequence'] = Sequence.get('account.batch.line.create')
-        lines = super(BatchLine, cls).create(vlist)
+        lines = super().create(vlist)
+        to_write = []
         for line in lines:
-            line.create_move()
+            sequence = Sequence.get('account.batch.line.create')
+            move = line.create_move()
+            move.save()
             if line.invoice:
-                line._add_payment_line()
+                line._add_payment_line(move)
                 # This check must be run *after* payment lines are added to
                 # have all the payment lines available
                 line._check_invoice_amount()
+            to_write.extend([
+                    [line], {'move': move.id, 'sequence': sequence},
+                    ])
+        if to_write:
+            with Transaction().set_context(skip_update_move=True):
+                cls.write(*to_write)
         return lines
 
     def create_move(self):
@@ -778,11 +779,6 @@ class BatchLine(ModelSQL, ModelView):
         key = {'date': self.date}
         move = self._get_move(key)
         move.lines = self.get_move_lines()
-        move.save()
-
-        self.write([self], {
-                'move': move.id,
-                })
         return move
 
     def _get_move(self, key):
@@ -837,11 +833,13 @@ class BatchLine(ModelSQL, ModelView):
             credit_account = self.contra_account
 
         cancel = self.is_cancelation_move
-
+        reverse_charge = False
         if self.tax:
             taxes = self._get_taxes(self.tax)
             base_tax_lines = []
             for tax in taxes:
+                if self.tax.reverse_charge:
+                    reverse_charge = True
                 total_amount = base_amount = amount
                 if not self.tax.reverse_charge:
                     base_amount = tax.reverse_compute(amount, [tax])
@@ -854,6 +852,11 @@ class BatchLine(ModelSQL, ModelView):
                 tax_amount = currency.round(tax_amount)
                 # Create the move line(s) for the taxes
                 # XXX: Unhandled case: tax_amount_second_currency
+                # Differentiate 3 cases of negative amounts:
+                # Expenses (purchases): reversed amount to other side
+                # Cancelations: (negative) amount to same side
+                # Taxes with negative rates (e.g. intracommunity): like
+                # expenses
                 tax_amount_second_currency = None
                 abs_tax_amount = abs(tax_amount)
                 abs_base_amount = abs(base_amount)
@@ -862,89 +865,73 @@ class BatchLine(ModelSQL, ModelView):
                 tax_debit_account = tax.credit_note_account
                 if not cancel:
                     if side == 'credit':
-                        tax_move_line = self._get_move_line(
-                            abs_tax_amount, _ZERO,
-                            tax_credit_account, second_currency,
-                            tax_amount_second_currency)
+                        tax_debit, tax_credit = -tax_amount, _ZERO
+                        tax_account = tax_credit_account
                     elif side == 'debit':
-                        tax_move_line = self._get_move_line(
-                            _ZERO, abs_tax_amount,
-                            tax_debit_account, second_currency,
-                            tax_amount_second_currency)
+                        tax_debit, tax_credit = _ZERO, tax_amount
+                        tax_account = tax_debit_account
                 elif cancel:
                     abs_tax_amount = -abs_tax_amount
                     abs_base_amount = -abs_base_amount
                     abs_total_amount = -abs_total_amount
                     if side == 'debit':
-                        tax_move_line = self._get_move_line(
-                            abs_tax_amount, _ZERO,
-                            tax_debit_account, second_currency,
-                            tax_amount_second_currency)
+                        tax_debit, tax_credit = -tax_amount, _ZERO
+                        tax_account = tax_debit_account
                     elif side == 'credit':
-                        tax_move_line = self._get_move_line(
-                            _ZERO, abs_tax_amount,
-                            tax_credit_account, second_currency,
-                            tax_amount_second_currency)
-                # Create the tax line(s) for the tax(es)
-                def get_tax_group(tax):
-                    tax_group = tax.group
-                    if not tax_group and tax.parent:
-                        tax_group = get_tax_group(tax.parent)
-                    return tax_group
+                        tax_debit, tax_credit = _ZERO, tax_amount
+                        tax_account = tax_credit_account
 
-                tax_group = get_tax_group(tax)
-                tax_group_kind = tax_group.kind if tax_group else None
-                if ((tax_group_kind is None
-                    or (tax_group_kind in ['sale', 'both'])
-                    and ((self.amount >= 0 and not cancel)
-                        or (self.amount < 0 and cancel)
-                                ))):
-                    #print('normal', tax_group_kind)
-                    base_tax_amount = abs_base_amount
-                else:
-                    #print('reverse', tax_group_kind)
-                    abs_tax_amount *= -1
-                    base_tax_amount = -abs_base_amount
+                tax_move_line = self._get_move_line(
+                    tax_debit, tax_credit, tax_account, second_currency,
+                    tax_amount_second_currency)
 
+                # Create the tax_line(s) for the tax(es)
                 tax_line = self._get_tax_line(abs_tax_amount, 'tax', tax)
                 tax_move_line.tax_lines += (tax_line,)
-                # Create the tax line(s) for the base
-                base_tax_line = self._get_tax_line(base_tax_amount,
-                    'base', tax)
-                base_tax_lines.append(base_tax_line)
+
+                # Create the tax_line(s) for the base
+                # Only one base line for all tax lines
+                if not base_tax_lines:
+                    base_tax_line = self._get_tax_line(abs_base_amount,
+                        'base', tax)
+                    base_tax_lines.append(base_tax_line)
                 res.append(tax_move_line)
 
             # Create the move line(s) for the base and total
             if not cancel:
                 if side == 'credit':
-                    base_move_line = self._get_move_line(
-                        abs_base_amount, _ZERO,
-                        debit_account, second_currency, amount_second_currency)
-                    total_move_line = self._get_move_line(
-                        _ZERO, abs_total_amount, credit_account,
-                        second_currency, amount_second_currency)
+                    base_debit, base_credit = -base_amount, _ZERO
+                    base_account = debit_account
+                    total_debit, total_credit = _ZERO, -total_amount
+                    total_account = credit_account
                 elif side == 'debit':
-                    base_move_line = self._get_move_line(
-                        _ZERO, abs_base_amount, credit_account,
-                        second_currency, amount_second_currency)
-                    total_move_line = self._get_move_line(
-                        abs_total_amount, _ZERO,
-                        debit_account, second_currency, amount_second_currency)
+                    base_debit, base_credit = _ZERO, base_amount
+                    base_account = credit_account
+                    total_debit, total_credit = total_amount, _ZERO
+                    total_account = debit_account
+                    if reverse_charge:
+                        base_account = debit_account
+                        total_account = credit_account
             elif cancel:
                 if side == 'debit':
-                    base_move_line = self._get_move_line(
-                        abs_base_amount, _ZERO, credit_account,
-                        second_currency, amount_second_currency)
-                    total_move_line = self._get_move_line(
-                        _ZERO, abs_total_amount,
-                        debit_account, second_currency, amount_second_currency)
+                    base_debit, base_credit = -base_amount, _ZERO
+                    base_account = credit_account
+                    total_debit, total_credit = _ZERO, -total_amount
+                    total_account = debit_account
                 elif side == 'credit':
-                    base_move_line = self._get_move_line(
-                        _ZERO, abs_base_amount,
-                        debit_account, second_currency, amount_second_currency)
-                    total_move_line = self._get_move_line(
-                        abs_total_amount, _ZERO, credit_account,
-                        second_currency, amount_second_currency)
+                    base_debit, base_credit = _ZERO, base_amount
+                    base_account = debit_account
+                    total_debit, total_credit = total_amount, _ZERO
+                    total_account = credit_account
+                    if reverse_charge:
+                        base_account = credit_account
+                        total_account = debit_account
+            base_move_line = self._get_move_line(
+                base_debit, base_credit, base_account, second_currency,
+                amount_second_currency)
+            total_move_line = self._get_move_line(
+                total_debit, total_credit, total_account, second_currency,
+                amount_second_currency)
             # add the previously created tax lines to the base
             base_move_line.tax_lines = base_tax_lines
             res.extend([base_move_line, total_move_line])
@@ -1021,28 +1008,25 @@ class BatchLine(ModelSQL, ModelView):
 
     @classmethod
     def write(cls, *args):
+        super().write(*args)
         actions = iter(args)
-        args = []
-
-        lines_to_update = set()
+        skip_update_move = Transaction().context.get('skip_update_move')
         for lines, values in zip(actions, actions):
-            if cls._check_update_move(set(values.keys())):
-                args.extend((lines, values))
-                lines_to_update.update(set(lines))
-        if lines_to_update:
-            super(BatchLine, cls).write(*args)
-            for line in lines_to_update:
-                if line.move:
-                    line._update_move_lines()
-                    if line.invoice:
-                        line._add_payment_line()
-                        # This check must be run *after* payment lines
-                        # are added to have all payment lines available
-                        line._check_invoice_amount()
+            if cls._check_update_move(values) and not skip_update_move:
+                for line in set(lines):
+                    if line.move:
+                        line._update_move_lines()
+                        if line.invoice:
+                            line._add_payment_line(line.move)
+                            # This check must be run *after* payment lines
+                            # are added to have all payment lines available
+                            line._check_invoice_amount()
 
     @classmethod
-    def _check_update_move(cls, fields):
-        if fields.difference(cls._check_modify_exclude):
+    def _check_update_move(cls, values):
+        fields = set(values.keys())
+        no_update_fields = set(cls._check_modify_exclude)
+        if fields.difference(no_update_fields):
             return True
         return False
 
@@ -1067,9 +1051,9 @@ class BatchLine(ModelSQL, ModelView):
         move.lines = self.get_move_lines()
         move.save()
 
-    def _add_payment_line(self):
+    def _add_payment_line(self, move):
         Invoice = Pool().get('account.invoice')
-        for move_line in self.move.lines:
+        for move_line in move.lines:
             if move_line.account == self.invoice.account:
                 Invoice.write([self.invoice], {
                         'payment_lines': [('add', [move_line.id])],
