@@ -9,7 +9,7 @@ from trytond.modules.account.tests.tools import get_accounts
 __all__ = ['create_tax', 'create_tax_code', 'create_tax_code_line']
 
 
-def create_tax(name, rate, company=None, config=None):
+def create_tax(name, rate, type='percentage', company=None, config=None):
     "Create a tax of rate"
     Tax = Model.get('account.tax', config=config)
 
@@ -21,8 +21,10 @@ def create_tax(name, rate, company=None, config=None):
     tax = Tax()
     tax.name = name
     tax.description = name
-    tax.type = 'percentage'
-    tax.rate = rate
+    if type is not None:
+        tax.type = type
+    if rate is not None:
+        tax.rate = rate
     tax.invoice_account = accounts['tax']
     tax.credit_note_account = accounts['tax']
     return tax
